const DEBUG=true
const APP_PORT=443
const SSL_CERT='certs/fullchain.cer';
const SSL_PRIVATEKEY='certs/rte.key';
const MQTT_SERVER='mqtt://127.0.0.1';
const HOSTNAME='gw.rt-eg.ru';
const MQTT_ID='yandexmqtt';
const YANDEX_USERID='testuser';
const CLIENT_ID='79a57bc8b4dd4affb5b7cd2fb1b6e409';//da76063c634b4fdc95a701579ff4647e';
const CLIENT_SECRET='11137b3728fe438b94cb70d81b89b0e7';//fd85b5252010433186e9ef78a712cbdb';
const SKILL_ID='7c6822ea-4280-4333-bb50-4c03a86f957e';
const DIALOGS_TOKEN='y0_AgAAAAAVCgBZAAT7owAAAAEJGOCjAABMiGk0oYdGwI5HknXK39BhvbTMOQ';  //Брать здесь: https://yandex.ru/dev/dialogs/smart-home/doc/ru/reference-alerts/resources-alerts
const AUTH_AVAILABLE=true;
const MANUFACTURER='wirenboard-unofficial';

const express = require("express");
const fs = require('fs');
const path = require('path');
const mqtt = require('mqtt');
var https = require('https');
var privateKey  = fs.readFileSync(SSL_PRIVATEKEY, 'utf8');
var certificate = fs.readFileSync(SSL_CERT, 'utf8');
var credentials = {key: privateKey, cert: certificate};
var yandexToken='';
var deviceStatus={};

// Запуск HTTPS-сервера
const app = express();
app.use(express.json());
var httpsServer = https.createServer(credentials, app);
httpsServer.listen(APP_PORT);

// Запуск MQTT-клиента
const mqttclient = mqtt.connect(MQTT_SERVER,{clientId:MQTT_ID});
const jsonParser = express.json();

// Загружаем сохранённый токен

function loadToken()
{
  if (!fs.existsSync('yandextoken.txt')) return;
  var data=fs.readFileSync('yandextoken.txt');
  yandexToken=JSON.parse(data).access_token;
  if (DEBUG) console.log('Loaded stored token: '+yandexToken);
}
// Полезные штуки

function isNumeric(n) {
  return !isNaN(parseFloat(n)) && isFinite(n);
}

function unixTimestamp () {  
  return Math.floor(Date.now() / 1000)
}

// Первичная авторизация в сервисе OAuth

async function proceedAuth(code)
{
    response = await fetch('https://oauth.yandex.ru/token',{
		method: 'POST',
		body: `grant_type=authorization_code&code=${code}&client_id=${CLIENT_ID}&client_secret=${CLIENT_SECRET}`
	    });

   data = await response.json();
   if (response.status!=200) { console.log(`Error: ${response.status}`); if (DEBUG) console.log(data); return false;}
  
   console.log('Got token: '+data.access_token);
   fs.writeFileSync('yandextoken.txt',JSON.stringify(data),(err) => {  if (err) console.log(err); });
   console.log('Tokens successfully stored to disk');

   return true;
}

// Формирование списка устройств

async function generateDevicelist(response)
{
    var deviceListString=``;
    var deviceList={};
    deviceList.request_id='';
    deviceList.payload={};
    deviceList.payload.user_id=YANDEX_USERID;
    deviceList.payload.devices=[];
    fs.readFile("static/devicelist.txt",'utf8', (err, data) => { if (err) throw err;
    ar=data.toString().split("\n");
    ar.forEach(ss=>{
      args=ss.split(";");
      if (args.length==5&&ss.substr(0,1)!='#')
      {
        var device={};
	deviceType='devices.types.'+args[1];
        deviceCapabilities=args[3];
        deviceProperties=args[4];
        device.id=args[0]; // MQTT-топик устройства
        device.name=args[2]; // Наименование устройства
        device.type=deviceType; // Тип устройства
        // Добавляем умения
        capsAr=deviceCapabilities.toString().split(',');
	device.capabilities=[];
        capsAr.forEach(cap=>{
	    if (cap.length>0)
	    {
	      curCap={};
	      capAr=cap.split(' ');
	      curCap.type='devices.capabilities.'+capAr[0];
	      curCap.parameters={};
 	      curCap.parameters.instance=capAr[1];
	      switch (capAr[0])
	      {
		  case 'range':
    	    		     curCap.parameters.range={};
			     curCap.parameters.unit='unit.'+capAr[2];
			     curCap.parameters.random_access=true;
			     curCap.parameters.range.min=parseFloat(capAr[3]);
			     curCap.parameters.range.max=parseFloat(capAr[4]);
			     curCap.parameters.range.precision=parseFloat(capAr[5]);
			     break;
		  case 'mode':
			     curCap.parameters.modes=[];
		             for (i=2;i<capAr.length;i++) curCap.parameters.modes.push({"value":capAr[i]});
			     break;
		}
	    curCap.retriviable=true; // Все свойства читаемые
	    device.capabilities.push(curCap);
	    }
	});
	device.device_info={};
	device.device_info.manufacturer=MANUFACTURER;
	device.device_info.model='Smart home';
	device.device_info.hw_version='1.0';
	device.device_info.sw_version='1.0';
	// Добавляем свойства
        propsAr=deviceProperties.toString().split(',');
	device.properties=[];
        propsAr.forEach(prop=>{
	    if (prop.length>0)
	    {
	      curProp={};
	      propAr=prop.split(' ');
	      propType=propAr[0];
	      curProp.type='devices.properties.'+propType;
	      curProp.retriveable=true;
	      curProp.reportable=true;
	      curProp.parameters={};
	      curProp.parameters.instance=propAr[1];
	      if (propType=='float') curProp.parameters.unit=propAr[2];
	       else
	      if (propType=='event')
		{
		    curProp.parameters.events=[];
		    for (i=2;i<propAr.length;i++) curProp.parameters.events.push({"value":propAr[i]});
		}
	      device.properties.push(curProp);
	    }
	});
	device.device_info={};
	device.device_info.manufacturer=MANUFACTURER;
	device.device_info.model='Smart home';
	device.device_info.hw_version='1.0';
	device.device_info.sw_version='1.0';
	// Добавляем устройство к результату
        deviceList.payload.devices.push(device);
      }
    });
    response.write(JSON.stringify(deviceList).toString());
    response.end();
  });
}

// Запрос статуса устройства

async function generateDeviceStatus(request,response)
{
    answer={};
    requestId=request.header('X-Request-Id');
    answer.request_id=requestId;
    answer.payload={};
    answer.payload.devices=[];
    fs.readFile("static/devicelist.txt",'utf8', (err, data) => { if (err) throw err;
      deviceListArray=data.toString().split("\n");
      deviceArray=request.body.devices;
      if (deviceArray)
      deviceArray.forEach(device=>{
  	id=device.id;
	if (DEBUG) console.log(`Requested device status for ${id}, request ID=${requestId}`);
        // Ищем устройство в файле с устройствами
           deviceListArray.forEach(ss=>{
             device={};
             device.id=id;
             args=ss.split(";");
             if (args.length==5&&ss.substr(0,1)!='#'&&args[0]==id)
             {
	       // Возвращаем информацию по умениям
	       device.capabilities=[];
	       capabilitiesLine=args[3];
               capabilitiesArray=capabilitiesLine.toString().split(',');
	       if(deviceStatus[id]&&deviceStatus[id].capabilities)
	         capabilitiesArray.forEach(cap=>{
 	           capAr=cap.split(' ');
		   capType=capAr[0];
		   instance=capAr[1];
		   capValue=deviceStatus[id]['capabilities'][instance];
		   if (typeof capValue!=="undefined")
		   {
		     curCap={};
		     curCap.state={};
		     curCap.type='devices.capabilities.'+capType;
		     curCap.state.instance=instance;
		     curCap.state.value=capValue;
    	             device.capabilities.push(curCap);
		   }
	         });
	       // Возвращаем информацию по свойствам
	       device.properties=[];
    	       propertiesLine=args[4];
               propertiesArray=propertiesLine.toString().split(',');
	       if(deviceStatus[id]&&deviceStatus[id].properties)
	         propertiesArray.forEach(prop=>{
 	           propAr=prop.split(' ');
		   propType=propAr[0];
		   instance=propAr[1];
		   propValue=deviceStatus[id]['properties'][instance];
		   if (typeof propValue!=="undefined")
		   {
		       curProp={};
		       curProp.type='devices.properties.'+propType;
		       curProp.state={};
		       curProp.state.instance=instance;
		       curProp.state.value=propValue;
    	               device.properties.push(curProp);
		   }
	         });

	       // ***
  	       answer.payload.devices.push(device);
	    }
          });
      });
      response.write(JSON.stringify(answer).toString());
      response.end();
   });
}

// Передаём действие в MQTT
async function generateAction(request,response)
{
    answer={};
    requestId=request.header('X-Request-Id');
    answer.request_id=requestId;
    answer.payload={};
    answer.payload.devices=[];
    if (request.body.payload)
     request.body.payload.devices.forEach(dvc=>{
       answerDevice={};
       answerDevice.id=dvc.id;
       answerDevice.capabilities=[];
       dvc.capabilities.forEach(requestCapability=>{
   	      if (DEBUG) console.log(`Requested device action for ${dvc.id}, request ID=${requestId}`);
	      answerCapability={};
	      answerCapability.type=requestCapability.type;
	      answerCapability.state={};
	      answerCapability.state.instance=requestCapability.state.instance;
	      answerCapability.state.action_result={}
	      answerCapability.state.action_result.status='DONE';
	      answerDevice.capabilities.push(answerCapability);
	      mqttclient.publish(`${MQTT_ID}/${dvc.id}/capabilities/${requestCapability.state.instance}`,requestCapability.state.value.toString());
	   }
	);
        answer.payload.devices.push(answerDevice);
     });
    response.write(JSON.stringify(answer).toString());
    response.end();
}

// Передача статусов обратно в Яндекс, инициированная MQTT

async function publishState(deviceId)
{
    answer={};
    answer.ts=unixTimestamp();
    answer.payload={};
    answer.payload.devices=[];
    answer.payload.user_id=YANDEX_USERID;
    id=deviceId;
    fs.readFile("static/devicelist.txt",'utf8', async (err, data) => { if (err) throw err;
      deviceListArray=data.toString().split("\n");
        // Ищем устройство в файле с устройствами
       deviceListArray.forEach(ss=>{
         device={};
         device.id=deviceId;
         args=ss.split(";");
         if (args.length==5&&ss.substr(0,1)!='#'&&args[0]==deviceId)
         {
	   // Возвращаем информацию по умениям
	   device.capabilities=[];
	   capabilitiesLine=args[3];
           capabilitiesArray=capabilitiesLine.toString().split(',');
	   if(deviceStatus[id]&&deviceStatus[id].capabilities)
	     capabilitiesArray.forEach(cap=>{
 	       capAr=cap.split(' ');
	       capType=capAr[0];
	       instance=capAr[1];
	       capValue=deviceStatus[id]['capabilities'][instance];
	       if (typeof capValue!=="undefined")
	       {
	         curCap={};
	         curCap.state={};
	         curCap.type='devices.capabilities.'+capType;
	         curCap.state.instance=instance;
	         curCap.state.value=capValue;
    	         device.capabilities.push(curCap);
	       }
	     });
	   // Возвращаем информацию по свойствам
	   device.properties=[];
    	   propertiesLine=args[4];
           propertiesArray=propertiesLine.toString().split(',');
	   if(deviceStatus[id]&&deviceStatus[id].properties)
	     propertiesArray.forEach(prop=>{
 	       propAr=prop.split(' ');
	       propType=propAr[0];
	       instance=propAr[1];
	       propValue=deviceStatus[id]['properties'][instance];
	       if (typeof propValue!=="undefined")
	       {
	         curProp={};
	         curProp.type='devices.properties.'+propType;
	         curProp.state={};
	         curProp.state.instance=instance;
	         curProp.state.value=propValue;
    	         device.properties.push(curProp);
	       }
	      });
	       // ***
  	       answer.payload.devices.push(device);
	 }
      });
      bodyData=JSON.stringify(answer).toString();
      response = await fetch(`https://dialogs.yandex.net/api/v1/skills/${SKILL_ID}/callback/state`,{
		method: 'POST',
		headers: {
			'Authorization': 'Bearer '+DIALOGS_TOKEN,
			'content-type': 'application/json'
		    },
		body: bodyData
	    });
      if (response.status!=202) { console.log(`Error: ${response.status}`);
            if (DEBUG)
	    {
        	data = await response.text();
		console.log(data);
    	    }
	}

   });

}

// Инициализация Web-сервисов

app.get("/proceedauth", function(request, response){
    if (!AUTH_AVAILABLE) { response.write('Auth disabled in config'); return; }
    code=request.query.code;
    if (DEBUG) console.log('Got temporary code: '+code);
    if (proceedAuth(code)) response.write('Success!'); else response.write('Failed');
    response.end();
});

// Регистрация в Яндекс-OAuth сервисе для получения токена, который потом будем проверять, чтобы убедиться, что команды идут от Яндекса, а не от злоумышленника
app.get("/auth", function(request, response){
    if (!AUTH_AVAILABLE) { response.write('Auth disabled in config'); return; }
    response.redirect(`https://oauth.yandex.ru/authorize?response_type=code&client_id=${CLIENT_ID}`);
});

// Яндекс проверяет, жив ли клиент
app.head("/v1.0", function(request, response){
    if (DEBUG) console.log('Requested keepalive');
    response.writeHead(200, {'Content-Type': 'text/html; charset=utf-8','Access-Control-Allow-Origin':'*'});
    response.end();
});

// Рассоединение эккаунтов - заглушка
app.post("/v1.0/user/unlink", function(request, response){
    if (DEBUG) console.log('Requested unlink');
    response.writeHead(200, {'Content-Type': 'text/html; charset=utf-8','Access-Control-Allow-Origin':'*'});
    answer={};
    answer.request_id=request.header('X-Request-Id');
    response.write(JSON.stringify(answer).toString());
    response.end();
});

// Яндекс запрашивает перечень устройств
app.get("/v1.0/user/devices", function(request, response){
    response.writeHead(200, {'Content-Type': 'text/html; charset=utf-8','Access-Control-Allow-Origin':'*'});
    if (DEBUG) console.log('Requested device list');
    if ((!DEBUG)&&request.headers.authorization.toString().substr(7)!=yandexToken) { console.log('Invalid token '+request.headers.authorization.toString().substr(7)); }
      else generateDevicelist(response);
});

// Яндекс запрашивает состояние устройства
app.post("/v1.0/user/devices/query", function(request, response){
    response.writeHead(200, {'Content-Type': 'text/html; charset=utf-8','Access-Control-Allow-Origin':'*'});
    if ((!DEBUG)&&request.headers.authorization.toString().substr(7)!=yandexToken) { console.log('Invalid token '+request.headers.authorization.toString().substr(7)); }
      else generateDeviceStatus(request,response);
});

// Яндекс передаёт действие
app.post("/v1.0/user/devices/action", function(request, response){
    response.writeHead(200, {'Content-Type': 'text/html; charset=utf-8','Access-Control-Allow-Origin':'*'});
    if ((!DEBUG)&&request.headers.authorization.toString().substr(7)!=yandexToken) { console.log('Invalid token '+request.headers.authorization.toString().substr(7)); }
      else generateAction(request,response);
});

// Инициализация MQTT

mqttclient.on('connect', () => {
  if (DEBUG) console.log(`MQTT connected: ${mqttclient.connected}`);
  if (mqttclient.connected === true) {
  mqttclient.subscribe(MQTT_ID+'/#');
}});

mqttclient.on('reconnect', function() {
    if (DEBUG) console.log('MQTT reconnected');
});

mqttclient.on('offline', function() {
    if (DEBUG) console.log('MQTT disconnected');
    process.exit(1);
});

mqttclient.on('close', function() {
    if (DEBUG) console.log('MQTT disconnected');
    process.exit(1);
});

mqttclient.on('error',(error) => {
        console.error(error);
        process.exit(1);
});

mqttclient.on('message',(topic, message) => {
     topicArr=topic.split('/');
     if (topicArr[0]==MQTT_ID&&topicArr.length>4&&(topicArr[2]=='capabilities'||topicArr[2]=='properties')&&topicArr[4]=='on') // пришло ко мне и есть устройство и свойство (4 уровня иерархии, крайний из которых - on)
     {
        deviceId=topicArr[1];
        statusTyp=topicArr[2];
	propertyName=topicArr[3];
	if (DEBUG) console.log(`MQTT received ${topic} is set to ${message}`);
	if (!deviceStatus[deviceId]) deviceStatus[deviceId]={};
	if (!deviceStatus[deviceId][statusTyp]) deviceStatus[deviceId][statusTyp]={};
	deviceStatus[deviceId][statusTyp][propertyName]=isNumeric(message)?parseFloat(message):(message=='true')?true:(message=='false')?false:message.toString();
	publishState(deviceId);
    }
});

// ***

loadToken();

console.log('Yandex2MQTT started');

